package ch.psi.daq.queryrest.config;

import java.util.Arrays;
import java.util.Set;
import java.util.stream.Collectors;

import javax.annotation.PostConstruct;
import javax.annotation.Resource;

import org.springframework.context.annotation.Configuration;
import org.springframework.context.annotation.PropertySource;
import org.springframework.core.env.Environment;
import org.springframework.web.servlet.config.annotation.CorsRegistry;
import org.springframework.web.servlet.config.annotation.WebMvcConfigurerAdapter;


@Configuration
@PropertySource(value = {"classpath:queryrest.properties"})
@PropertySource(value = {"file:${user.home}/.config/daq/queryrest.properties"}, ignoreResourceNotFound = true)
public class QueryRestConfigCORS extends WebMvcConfigurerAdapter {

   @Resource
   private Environment env;
   
   @Resource(name = QueryRestConfig.BEAN_NAME_CORS_ALLOWEDORIGINS)
   private String configuredOrigins;

   @Resource(name = QueryRestConfig.BEAN_NAME_CORS_FORCEALLHEADERS)
   private Boolean forceAllHeaders;

   private String[] allowedOrigins;

   @PostConstruct
   public void afterPropertiesSet() {
      Set<String> origs = Arrays.stream(configuredOrigins.split(",")).map(s -> s.trim()).collect(Collectors.toSet());
      allowedOrigins = origs.toArray(new String[origs.size()]);
   }

   @Override
   public void addCorsMappings(CorsRegistry registry) {
      
      boolean corsEnabled = Boolean.valueOf(env.getProperty("queryrest.cors.enable", "false"));
      boolean forceAll = Boolean.valueOf(env.getProperty("queryrest.cors.forceallheaders", "false"));

      if (corsEnabled) {
         
         if (forceAll) {
            registry
            .addMapping("/**")
            .allowedOrigins("*")
            .allowedMethods("GET", "POST", "PUT", "DELETE", "OPTIONS")
            .allowedHeaders("Origin", "Accept", "X-Requested-With", "Content-Type", "Access-Control-Request-Method",
                  "Access-Control-Request-Headers")
            .allowCredentials(true)
            .maxAge(1800);
         } else if (!forceAll) {
            // see https://spring.io/blog/2015/06/08/cors-support-in-spring-framework
            registry
            .addMapping("/**")
            .allowedOrigins(allowedOrigins)
            .allowedMethods("GET", "POST", "PUT", "DELETE", "OPTIONS")
            .allowedHeaders("Origin", "Accept", "X-Requested-With", "Content-Type", "Access-Control-Request-Method",
                  "Access-Control-Request-Headers")
            .allowCredentials(true)
            .maxAge(1800);
         }
      }
   }
}
