package ch.psi.daq.queryrest.response.csv;

import java.io.OutputStream;
import java.util.List;
import java.util.Map.Entry;
import java.util.stream.Stream;

import javax.servlet.http.HttpServletResponse;

import org.apache.commons.lang3.tuple.Triple;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.context.ApplicationContext;

import com.hazelcast.util.collection.ArrayUtils;

import ch.psi.daq.domain.FieldNames;
import ch.psi.daq.domain.json.ChannelName;
import ch.psi.daq.domain.query.DAQQueries;
import ch.psi.daq.domain.query.DAQQueryElement;
import ch.psi.daq.domain.query.operation.AggregationType;
import ch.psi.daq.domain.query.operation.Compression;
import ch.psi.daq.domain.query.operation.QueryField;
import ch.psi.daq.domain.query.operation.ResponseFormat;
import ch.psi.daq.query.model.impl.BackendQuery;
import ch.psi.daq.queryrest.query.QueryManager;
import ch.psi.daq.queryrest.response.AbstractHTTPResponse;

public class CSVHTTPResponse extends AbstractHTTPResponse {
   private static final Logger LOGGER = LoggerFactory.getLogger(CSVHTTPResponse.class);

   public static final String FORMAT = "csv";
   public static final String CONTENT_TYPE = "text/csv";

   public CSVHTTPResponse() {
      super(ResponseFormat.CSV);
   }

   public CSVHTTPResponse(Compression compression) {
      this();
      setCompression(compression);
   }

   @Override
   public void respond(ApplicationContext context, DAQQueries queries, HttpServletResponse httpResponse) throws Exception {
      OutputStream out = handleCompressionAndResponseHeaders(httpResponse, CONTENT_TYPE);

      // do csv specific validations
      validateQueries(queries);

      try {
         LOGGER.debug("Executing query '{}'", queries);

         QueryManager queryManager = context.getBean(QueryManager.class);
         CSVResponseStreamWriter streamWriter = context.getBean(CSVResponseStreamWriter.class);

         // execute query
         List<Entry<DAQQueryElement, Stream<Triple<BackendQuery, ChannelName, ?>>>> result =
               queryManager.executeQueries(queries);
         // write the response back to the client using java 8 streams
         streamWriter.respond(result, out);
      } catch (Exception e) {
         LOGGER.error("Failed to execute query '{}'.", queries, e);
         throw e;
      }
   }

   protected void validateQueries(DAQQueries queries) {
      for (DAQQueryElement query : queries) {
         if (!(query.getAggregation() == null || AggregationType.value.equals(query.getAggregation().getAggregationType()))) {
            // We allow only no aggregation or value aggregation as
            // extrema: nested structure and not clear how to map it to one line
            // index: value is an array of Statistics whose size is not clear at initialization time
            String message = "CSV export does not support '" + query.getAggregation().getAggregationType() + "'";
            LOGGER.warn(message);
            throw new IllegalArgumentException(message);
         }


         if (!ArrayUtils.contains(query.getColumns(), FieldNames.FIELD_GLOBAL_TIME)) {
            query.addField(QueryField.globalMillis);
         }
      }
   }
}
