package ch.psi.daq.test.queryrest.controller;

import static org.junit.Assert.assertFalse;

import org.junit.After;
import org.junit.Test;
import org.springframework.http.MediaType;
import org.springframework.test.web.servlet.MvcResult;
import org.springframework.test.web.servlet.request.MockMvcRequestBuilders;
import org.springframework.test.web.servlet.result.MockMvcResultHandlers;
import org.springframework.test.web.servlet.result.MockMvcResultMatchers;

import ch.psi.daq.common.time.TimeUtils;
import ch.psi.daq.domain.json.ChannelName;
import ch.psi.daq.domain.query.status.channel.ChannelStatusQuery;
import ch.psi.daq.domain.reader.Backend;
import ch.psi.daq.queryrest.controller.QueryRestController;
import ch.psi.daq.test.queryrest.AbstractDaqRestTest;

/**
 * Tests the {@link DaqController} implementation.
 */
public class QueryRestControllerChannelStatusTest extends AbstractDaqRestTest {

   @After
   public void tearDown() throws Exception {}

   @Test
   public void testChannelStatusQuery_01() throws Exception {
      ChannelStatusQuery query = new ChannelStatusQuery("DataBuffer1", "DataBuffer2");

      String content = mapper.writeValueAsString(query);
      System.out.println(content);

      this.mockMvc
            .perform(MockMvcRequestBuilders
                  .post(QueryRestController.PATH_STATUS_CHANNELS)
                  .contentType(MediaType.APPLICATION_JSON)
                  .content(content))

            .andExpect(MockMvcResultMatchers.status().isOk())
            .andExpect(MockMvcResultMatchers.jsonPath("$").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].channel.backend").value(Backend.SF_DATABUFFER.getKey()))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].channel.name").value("DataBuffer1"))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].recording").value(false))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].connected").value(false))
            .andExpect(
                  MockMvcResultMatchers.jsonPath("$[0].lastEventDate").value(
                        TimeUtils.format(TimeUtils.getTimeFromMillis(0, 0))))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].channel.backend").value(Backend.SF_DATABUFFER.getKey()))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].channel.name").value("DataBuffer2"))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].recording").value(false))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].connected").value(false))
            .andExpect(
                  MockMvcResultMatchers.jsonPath("$[1].lastEventDate").value(
                        TimeUtils.format(TimeUtils.getTimeFromMillis(0, 0))))
            .andExpect(MockMvcResultMatchers.jsonPath("$[2]").doesNotExist());
   }

   @Test
   public void testChannelStatusQuery_02() throws Exception {
      ChannelStatusQuery query = new ChannelStatusQuery();
      query.addChannel(new ChannelName("ArchiverAppliance1", Backend.SF_ARCHIVERAPPLIANCE));
      query.addChannel(new ChannelName("ArchiverAppliance2", Backend.SF_ARCHIVERAPPLIANCE));

      String content = mapper.writeValueAsString(query);
      System.out.println(content);

      this.mockMvc
            .perform(MockMvcRequestBuilders
                  .post(QueryRestController.PATH_STATUS_CHANNELS)
                  .contentType(MediaType.APPLICATION_JSON)
                  .content(content))

            .andExpect(MockMvcResultMatchers.status().isOk())
            .andExpect(MockMvcResultMatchers.jsonPath("$").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0]").exists())
            .andExpect(
                  MockMvcResultMatchers.jsonPath("$[0].channel.backend").value(Backend.SF_ARCHIVERAPPLIANCE.getKey()))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].channel.name").value("ArchiverAppliance1"))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].recording").value(true))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].connected").value(true))
            .andExpect(
                  MockMvcResultMatchers.jsonPath("$[0].lastEventDate").value(
                        TimeUtils.format(TimeUtils.getTimeFromMillis(1467638000000L, 0))))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1]").exists())
            .andExpect(
                  MockMvcResultMatchers.jsonPath("$[1].channel.backend").value(Backend.SF_ARCHIVERAPPLIANCE.getKey()))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].channel.name").value("ArchiverAppliance2"))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].recording").value(true))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].connected").value(true))
            .andExpect(
                  MockMvcResultMatchers.jsonPath("$[1].lastEventDate").value(
                        TimeUtils.format(TimeUtils.getTimeFromMillis(1467638000000L, 0))))
            .andExpect(MockMvcResultMatchers.jsonPath("$[2]").doesNotExist());
   }

   @Test
   public void testChannelStatusQuery_03() throws Exception {
      ChannelStatusQuery query = new ChannelStatusQuery();
      query.addChannel(new ChannelName("ArchiverAppliance1", Backend.SF_ARCHIVERAPPLIANCE));
      query.addChannel(new ChannelName("ArchiverAppliance2", Backend.SF_ARCHIVERAPPLIANCE));

      String content = mapper.writeValueAsString(query);
      System.out.println(content);

      MvcResult result = this.mockMvc.perform(MockMvcRequestBuilders
            .post(QueryRestController.PATH_STATUS_CHANNELS)
            .contentType(MediaType.APPLICATION_JSON)
            .content(content))
            .andDo(MockMvcResultHandlers.print())
            .andExpect(MockMvcResultMatchers.status().isOk())
            .andReturn();

      String response = result.getResponse().getContentAsString();
      System.out.println("Response: " + response);

      assertFalse(response.contains("lastEventTime"));
   }
}
