package ch.psi.daq.test.queryrest.controller;

import java.util.Arrays;

import org.junit.After;
import org.junit.Test;
import org.springframework.http.MediaType;
import org.springframework.test.web.servlet.request.MockMvcRequestBuilders;
import org.springframework.test.web.servlet.result.MockMvcResultHandlers;
import org.springframework.test.web.servlet.result.MockMvcResultMatchers;
import org.springframework.test.web.servlet.setup.MockMvcBuilders;

import ch.psi.daq.common.ordering.Ordering;
import ch.psi.daq.common.time.TimeUtils;
import ch.psi.daq.domain.json.ChannelName;
import ch.psi.daq.domain.query.DAQQueries;
import ch.psi.daq.domain.query.DAQQuery;
import ch.psi.daq.domain.query.DAQQueryElement;
import ch.psi.daq.domain.query.channels.ChannelsRequest;
import ch.psi.daq.domain.query.operation.Aggregation;
import ch.psi.daq.domain.query.operation.AggregationDescriptor;
import ch.psi.daq.domain.query.operation.AggregationType;
import ch.psi.daq.domain.query.operation.Compression;
import ch.psi.daq.domain.query.operation.Extrema;
import ch.psi.daq.domain.query.operation.QueryField;
import ch.psi.daq.domain.reader.Backend;
import ch.psi.daq.domain.request.range.RequestRangeDate;
import ch.psi.daq.domain.request.range.RequestRangePulseId;
import ch.psi.daq.domain.request.range.RequestRangeTime;
import ch.psi.daq.domain.test.TestTimeUtils;
import ch.psi.daq.queryrest.controller.QueryRestController;
import ch.psi.daq.queryrest.response.json.JSONHTTPResponse;
import ch.psi.daq.test.queryrest.AbstractDaqRestTest;

/**
 * Tests the {@link DaqController} implementation.
 */
public class QueryRestControllerJsonTest extends AbstractDaqRestTest {

   public static final String TEST_CHANNEL_01 = "testChannel1";
   public static final String TEST_CHANNEL_02 = "testChannel2";
   public static final String TEST_CHANNEL_WAVEFORM_01 = "testChannelWaveform1";
   public static final String[] TEST_CHANNEL_NAMES = new String[] {TEST_CHANNEL_01, TEST_CHANNEL_02};

   @After
   public void tearDown() throws Exception {}

   @Test
   public void testChannelNameQuery() throws Exception {

      this.mockMvc
            .perform(
                  MockMvcRequestBuilders
                        .get(QueryRestController.PATH_CHANNELS)
                        .contentType(MediaType.APPLICATION_JSON))
            .andDo(MockMvcResultHandlers.print())
            .andExpect(MockMvcResultMatchers.status().isOk())
            .andExpect(MockMvcResultMatchers.jsonPath("$").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].backend").value(Backend.SF_DATABUFFER.getKey()))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].channels").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].channels[0]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].channels[0]").value("BoolScalar"))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].channels[1]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].channels[1]").value("BoolWaveform"))
      // .andExpect(MockMvcResultMatchers.jsonPath("$[1]").exists())
      // .andExpect(MockMvcResultMatchers.jsonPath("$[1].backend").value(Backend.archiverappliance.name()))
      // .andExpect(MockMvcResultMatchers.jsonPath("$[1].channels").isArray())
      // .andExpect(MockMvcResultMatchers.jsonPath("$[1].channels[0]").exists())
      // .andExpect(
      // MockMvcResultMatchers.jsonPath("$[1].channels[0]").value(DummyArchiverApplianceReader.TEST_CHANNEL_1))
      // .andExpect(MockMvcResultMatchers.jsonPath("$[1].channels[1]").exists())
      // .andExpect(
      // MockMvcResultMatchers.jsonPath("$[1].channels[1]").value(DummyArchiverApplianceReader.TEST_CHANNEL_2))
      ;

   }

   @Test
   public void testSpecificChannelSearch() throws Exception {
      this.mockMvc.perform(
            MockMvcRequestBuilders
                  .get(QueryRestController.PATH_CHANNELS + "/int32")
                  .contentType(MediaType.APPLICATION_JSON))
            .andDo(MockMvcResultHandlers.print())
            .andExpect(MockMvcResultMatchers.status().isOk())
            .andExpect(MockMvcResultMatchers.jsonPath("$").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].backend").value(Backend.SF_DATABUFFER.getKey()))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].channels").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].channels[0]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].channels[0]").value("Int32Scalar"))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].channels[1]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].channels[1]").value("Int32Waveform"))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].channels[2]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].channels[2]").value("UInt32Scalar"))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].channels[3]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].channels[3]").value("UInt32Waveform"))
      // .andExpect(MockMvcResultMatchers.jsonPath("$[1]").exists())
      // .andExpect(MockMvcResultMatchers.jsonPath("$[1].backend").value(Backend.archiverappliance.name()))
      // .andExpect(MockMvcResultMatchers.jsonPath("$[1].channels").isArray())
      // .andExpect(MockMvcResultMatchers.jsonPath("$[1].channels[0]").doesNotExist())
      ;
   }

   @Test
   public void testChannelNameQueryBackendOrder() throws Exception {
      ChannelsRequest request = new ChannelsRequest("int64", Ordering.desc, Backend.SF_DATABUFFER);

      String content = mapper.writeValueAsString(request);
      System.out.println(content);

      this.mockMvc
            .perform(MockMvcRequestBuilders
                  .post(QueryRestController.PATH_CHANNELS)
                  .contentType(MediaType.APPLICATION_JSON)
                  .content(content))

            .andExpect(MockMvcResultMatchers.status().isOk())
            .andExpect(MockMvcResultMatchers.jsonPath("$").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].backend").value(Backend.SF_DATABUFFER.getKey()))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].channels").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].channels[0]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].channels[0]").value("UInt64Waveform"))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].channels[1]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].channels[1]").value("UInt64Scalar"))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].channels[2]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].channels[2]").value("Int64Waveform"))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].channels[3]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].channels[3]").value("Int64Scalar"))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1]").doesNotExist());
   }

   @Test
   public void testChannelNameQueryReload() throws Exception {
      ChannelsRequest request = new ChannelsRequest();

      String content = mapper.writeValueAsString(request);
      System.out.println(content);

      this.mockMvc
            .perform(MockMvcRequestBuilders
                  .post(QueryRestController.PATH_CHANNELS)
                  .contentType(MediaType.APPLICATION_JSON)
                  .content(content))

            .andExpect(MockMvcResultMatchers.status().isOk())
            .andExpect(MockMvcResultMatchers.jsonPath("$").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].backend").value(Backend.SF_DATABUFFER.getKey()))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].channels").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].channels[24]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].channels[25]").doesNotExist())
      // .andExpect(MockMvcResultMatchers.jsonPath("$[1]").exists())
      // .andExpect(MockMvcResultMatchers.jsonPath("$[1].backend").value(Backend.archiverappliance.name()))
      // .andExpect(MockMvcResultMatchers.jsonPath("$[1].channels").isArray())
      // .andExpect(MockMvcResultMatchers.jsonPath("$[1].channels[2]").exists())
      // .andExpect(MockMvcResultMatchers.jsonPath("$[1].channels[3]").doesNotExist())
      ;

      // each reload add another channel
      request.setReload(true);

      content = mapper.writeValueAsString(request);
      System.out.println(content);

      this.mockMvc
            .perform(MockMvcRequestBuilders
                  .post(QueryRestController.PATH_CHANNELS)
                  .contentType(MediaType.APPLICATION_JSON)
                  .content(content))

            .andExpect(MockMvcResultMatchers.status().isOk())
            .andExpect(MockMvcResultMatchers.jsonPath("$").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].backend").value(Backend.SF_DATABUFFER.getKey()))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].channels").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].channels[25]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].channels[26]").doesNotExist())
      // .andExpect(MockMvcResultMatchers.jsonPath("$[1]").exists())
      // .andExpect(MockMvcResultMatchers.jsonPath("$[1].backend").value(Backend.archiverappliance.name()))
      // .andExpect(MockMvcResultMatchers.jsonPath("$[1].channels").isArray())
      // .andExpect(MockMvcResultMatchers.jsonPath("$[1].channels[3]").exists())
      // .andExpect(MockMvcResultMatchers.jsonPath("$[1].channels[4]").doesNotExist())
      ;
   }

   @Test
   public void testCorsFilterNoHeaders() throws Exception {
      this.mockMvc = MockMvcBuilders.webAppContextSetup(webApplicationContext).build();

      this.mockMvc.perform(
            MockMvcRequestBuilders
                  .options(QueryRestController.PATH_CHANNELS)
                  .contentType(MediaType.APPLICATION_JSON))
            .andDo(MockMvcResultHandlers.print())
            .andExpect(MockMvcResultMatchers.status().isOk())
            // we didn't set the 'Origin' header so no access-control
            .andExpect(MockMvcResultMatchers.header().doesNotExist("Access-Control-Allow-Origin"));
   }

   @Test
   public void testCorsFilterIncludesHeaders() throws Exception {
      // all headers are set
      this.mockMvc = MockMvcBuilders.webAppContextSetup(webApplicationContext).build();

      // curl -H "Origin: *" -H "Access-Control-Request-Method: POST" -X OPTIONS -v
      // http://localhost:8080/channels
      this.mockMvc.perform(
            MockMvcRequestBuilders
                  .options(QueryRestController.PATH_CHANNELS)
                  .header("Origin", "*")
                  .header("Access-Control-Request-Method", "POST")
                  .contentType(MediaType.APPLICATION_JSON))
            .andDo(MockMvcResultHandlers.print())
            .andExpect(MockMvcResultMatchers.status().isOk())
            .andExpect(MockMvcResultMatchers.header().string("Access-Control-Allow-Origin", "*"));

      // curl -H "Origin: http://localhost:8080" -H "Access-Control-Request-Method: POST" -X OPTIONS
      // -v http://localhost:8080/channels
      this.mockMvc.perform(
            MockMvcRequestBuilders
                  .options(QueryRestController.PATH_CHANNELS)
                  .header("Origin", "http://localhost:8080")
                  .header("Access-Control-Request-Method", "POST")
                  .contentType(MediaType.APPLICATION_JSON))
            .andDo(MockMvcResultHandlers.print())
            .andExpect(MockMvcResultMatchers.status().isOk())
            .andExpect(MockMvcResultMatchers.header().string("Access-Control-Allow-Origin", "http://localhost:8080"));
   }

   @Test
   public void testPulseRangeQuery() throws Exception {
      DAQQuery request = new DAQQuery(
            new RequestRangePulseId(
                  100,
                  101),
            TEST_CHANNEL_NAMES);
      request.addField(QueryField.pulseId);
      request.addField(QueryField.globalSeconds);
      request.addField(QueryField.globalMillis);
      request.addField(QueryField.iocSeconds);
      request.addField(QueryField.iocMillis);

      String content = mapper.writeValueAsString(request);
      System.out.println(content);

      this.mockMvc
            .perform(MockMvcRequestBuilders
                  .post(QueryRestController.PATH_QUERY)
                  .contentType(MediaType.APPLICATION_JSON)
                  .content(content))

            .andDo(MockMvcResultHandlers.print())
            .andExpect(MockMvcResultMatchers.status().isOk())
            .andExpect(MockMvcResultMatchers.jsonPath("$").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].channel.name").value(TEST_CHANNEL_01))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].pulseId").value(100))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].globalSeconds").value(
                  TestTimeUtils.getTimeStr(1, 0)))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].globalMillis").value(1000))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].iocSeconds").value(
                  TestTimeUtils.getTimeStr(1, 0)))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].iocMillis").value(1000))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].pulseId").value(101))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].globalSeconds").value(
                  TestTimeUtils.getTimeStr(1, 10000000)))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].globalMillis").value(1010))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].iocSeconds").value(
                  TestTimeUtils.getTimeStr(1, 10000000)))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].iocMillis").value(1010))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].channel.name").value(TEST_CHANNEL_02))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data[0].pulseId").value(100))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data[0].globalSeconds").value(
                  TestTimeUtils.getTimeStr(1, 0)))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data[0].globalMillis").value(1000))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data[0].iocSeconds").value(
                  TestTimeUtils.getTimeStr(1, 0)))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data[0].iocMillis").value(1000))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data[1].pulseId").value(101))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data[1].globalSeconds").value(
                  TestTimeUtils.getTimeStr(1, 10000000)))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data[1].globalMillis").value(1010))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data[1].iocSeconds").value(
                  TestTimeUtils.getTimeStr(1, 10000000)))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data[1].iocMillis").value(1010));
   }

   @Test
   public void testPulseRangeQuery_Fields() throws Exception {
      DAQQuery request = new DAQQuery(
            new RequestRangePulseId(
                  100,
                  199),
            new AggregationDescriptor().setNrOfBins(2),
            TEST_CHANNEL_NAMES);
      request.addField(QueryField.pulseId);
      request.addField(QueryField.eventCount);

      String content = mapper.writeValueAsString(request);
      System.out.println(content);

      this.mockMvc
            .perform(MockMvcRequestBuilders
                  .post(QueryRestController.PATH_QUERY)
                  .contentType(MediaType.APPLICATION_JSON)
                  .content(content))

            .andDo(MockMvcResultHandlers.print())
            .andExpect(MockMvcResultMatchers.status().isOk())
            .andExpect(MockMvcResultMatchers.jsonPath("$").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].channel.name").value(TEST_CHANNEL_01))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].pulseId").value(100))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].eventCount").value(50))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].pulseId").value(150))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].eventCount").value(50))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].channel.name").value(TEST_CHANNEL_02))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data[0].pulseId").value(100))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data[0].eventCount").value(50))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data[1].pulseId").value(150))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data[1].eventCount").value(50));
   }

   @Test
   public void testPulseRangeQueryBackends() throws Exception {
      DAQQuery request = new DAQQuery(
            new RequestRangePulseId(
                  100,
                  101),
            new ChannelName(TEST_CHANNEL_01, Backend.SF_DATABUFFER),
            new ChannelName(TEST_CHANNEL_02, Backend.SF_ARCHIVERAPPLIANCE));

      String content = mapper.writeValueAsString(request);
      System.out.println(content);

      this.mockMvc
            .perform(MockMvcRequestBuilders
                  .post(QueryRestController.PATH_QUERY)
                  .contentType(MediaType.APPLICATION_JSON)
                  .content(content))

            .andDo(MockMvcResultHandlers.print())
            .andExpect(MockMvcResultMatchers.status().isOk())
            .andExpect(MockMvcResultMatchers.jsonPath("$").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].channel").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].channel.name").value(TEST_CHANNEL_01))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].pulseId").value(100))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].globalSeconds").value(
                  TestTimeUtils.getTimeStr(1, 0)))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].pulseId").value(101))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].globalSeconds").value(
                  TestTimeUtils.getTimeStr(1, 10000000)))
      // .andExpect(MockMvcResultMatchers.jsonPath("$[1]").exists())
      // .andExpect(MockMvcResultMatchers.jsonPath("$[1].channel").isMap())
      // .andExpect(MockMvcResultMatchers.jsonPath("$[1].channel.name").value(TEST_CHANNEL_02))
      // .andExpect(MockMvcResultMatchers.jsonPath("$[1].data").isArray())
      // .andExpect(MockMvcResultMatchers.jsonPath("$[1].data[0].pulseId").value(10))
      // .andExpect(MockMvcResultMatchers.jsonPath("$[1].data[0].globalMillis").value(100))
      // .andExpect(MockMvcResultMatchers.jsonPath("$[1].data[1].pulseId").value(11))
      // .andExpect(MockMvcResultMatchers.jsonPath("$[1].data[1].globalMillis").value(110))
      ;
   }

   @Test
   public void testPulseRangeQueries() throws Exception {
      String testChannel3 = "testChannel3";
      DAQQueries request = new DAQQueries(
            new DAQQueryElement(
                  new RequestRangePulseId(
                        100,
                        101),
                  TEST_CHANNEL_NAMES),
            new DAQQueryElement(
                  new RequestRangePulseId(
                        100,
                        101),
                  testChannel3));

      String content = mapper.writeValueAsString(request);
      System.out.println(content);

      this.mockMvc
            .perform(MockMvcRequestBuilders
                  .post(QueryRestController.PATH_QUERIES)
                  .contentType(MediaType.APPLICATION_JSON)
                  .content(content))

            .andDo(MockMvcResultHandlers.print())
            .andExpect(MockMvcResultMatchers.status().isOk())
            .andExpect(MockMvcResultMatchers.jsonPath("$").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0]").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0][0]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0][0].channel.name").value(TEST_CHANNEL_01))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0][0].data[0].pulseId").value(100))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0][0].data[0].globalSeconds").value(
                  TestTimeUtils.getTimeStr(1, 0)))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0][0].data[1].pulseId").value(101))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0][0].data[1].globalSeconds").value(
                  TestTimeUtils.getTimeStr(1, 10000000)))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0][1]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0][1].channel.name").value(TEST_CHANNEL_02))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0][1].data").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0][1].data[0].pulseId").value(100))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0][1].data[0].globalSeconds").value(
                  TestTimeUtils.getTimeStr(1, 0)))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0][1].data[1].pulseId").value(101))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0][1].data[1].globalSeconds").value(
                  TestTimeUtils.getTimeStr(1, 10000000)))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$[1]").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$[1][0]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$[1][0].channel.name").value(testChannel3))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1][0].data[0].pulseId").value(100))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1][0].data[0].globalSeconds").value(
                  TestTimeUtils.getTimeStr(1, 0)))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1][0].data[1].pulseId").value(101))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1][0].data[1].globalSeconds").value(
                  TestTimeUtils.getTimeStr(1, 10000000)));
   }

   @Test
   public void testTimeRangeQuery() throws Exception {
      DAQQuery request = new DAQQuery(
            new RequestRangeTime(
                  TimeUtils.getTimeFromMillis(2000, 0),
                  TimeUtils.getTimeFromMillis(2010, 0)),
            TEST_CHANNEL_NAMES);

      String content = mapper.writeValueAsString(request);

      this.mockMvc.perform(MockMvcRequestBuilders
            .post(QueryRestController.PATH_QUERY)
            .contentType(MediaType.APPLICATION_JSON)
            .content(content))

            .andDo(MockMvcResultHandlers.print())
            .andExpect(MockMvcResultMatchers.status().isOk())
            .andExpect(MockMvcResultMatchers.jsonPath("$").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].channel").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].channel.name").value(TEST_CHANNEL_01))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].pulseId").value(200))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].globalSeconds").value(
                  TestTimeUtils.getTimeStr(2, 0)))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].pulseId").value(201))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].globalSeconds").value(
                  TestTimeUtils.getTimeStr(2, 10000000)))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].channel").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].channel.name").value(TEST_CHANNEL_02))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data[0].pulseId").value(200))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data[0].globalSeconds").value(
                  TestTimeUtils.getTimeStr(2, 0)))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data[1].pulseId").value(201))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data[1].globalSeconds").value(
                  TestTimeUtils.getTimeStr(2, 10000000)));
   }

   @Test
   public void testDateRangeQuery() throws Exception {
      String startDate = TimeUtils.format(1000);
      String endDate = TimeUtils.format(1010);
      DAQQuery request = new DAQQuery(
            new RequestRangeDate(
                  startDate,
                  endDate),
            TEST_CHANNEL_NAMES);

      String content = mapper.writeValueAsString(request);
      System.out.println(content);

      this.mockMvc
            .perform(
                  MockMvcRequestBuilders
                        .post(QueryRestController.PATH_QUERY)
                        .contentType(MediaType.APPLICATION_JSON)
                        .content(content)
            )
            .andDo(MockMvcResultHandlers.print())
            .andExpect(MockMvcResultMatchers.status().isOk())
            .andExpect(MockMvcResultMatchers.jsonPath("$").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].channel.name").value(TEST_CHANNEL_01))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].channel.backend").value(Backend.SF_DATABUFFER.getKey()))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].pulseId").value(100))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].globalSeconds").value(
                  TestTimeUtils.getTimeStr(1, 0)))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].pulseId").value(101))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].globalSeconds").value(
                  TestTimeUtils.getTimeStr(1, 10000000)))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].channel.name").value(TEST_CHANNEL_02))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].channel.backend").value("sf-databuffer"))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data[0].pulseId").value(100))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data[0].globalSeconds").value(
                  TestTimeUtils.getTimeStr(1, 0)))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data[1].pulseId").value(101))
            .andExpect(MockMvcResultMatchers.jsonPath("$[1].data[1].globalSeconds").value(
                  TestTimeUtils.getTimeStr(1, 10000000)));
   }

   @Test
   public void testExtremaAggregation() throws Exception {
      DAQQuery request = new DAQQuery(
            new RequestRangePulseId(
                  100,
                  101),
            Ordering.asc,
            new AggregationDescriptor(AggregationType.extrema),
            TEST_CHANNEL_NAMES[0]);

      String content = mapper.writeValueAsString(request);

      this.mockMvc
            .perform(MockMvcRequestBuilders.post(QueryRestController.PATH_QUERY)
                  .contentType(MediaType.APPLICATION_JSON)
                  .content(content))

            .andDo(MockMvcResultHandlers.print())
            .andExpect(MockMvcResultMatchers.status().isOk())

            .andExpect(MockMvcResultMatchers.jsonPath("$").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].channel").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].channel.name").value(TEST_CHANNEL_NAMES[0]))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data.minima").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data.minima.min").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data.minima.min.value").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data.minima.min.value").value(Double.valueOf(100)))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data.minima.min.event").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data.minima.min.event.pulseId").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data.minima.min.event.pulseId").value(100))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data.maxima").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data.maxima.max").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data.maxima.max.value").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data.maxima.max.value").value(Double.valueOf(101)))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data.maxima.max.event").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data.maxima.max.event.pulseId").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data.maxima.max.event.pulseId").value(101));
   }

   @Test
   public void testDateRangeQueryNrOfBinsAggregate() throws Exception {
      long startTime = 1000;
      long endTime = 1099;
      String startDate = TimeUtils.format(startTime);
      String endDate = TimeUtils.format(endTime);
      DAQQuery request =
            new DAQQuery(
                  new RequestRangeDate(
                        startDate,
                        endDate),
                  new AggregationDescriptor().setNrOfBins(2).setAggregations(
                        Arrays.asList(Aggregation.min, Aggregation.mean, Aggregation.max)),
                  TEST_CHANNEL_01);

      String content = mapper.writeValueAsString(request);
      System.out.println(content);

      this.mockMvc
            .perform(
                  MockMvcRequestBuilders
                        .post(QueryRestController.PATH_QUERY)
                        .contentType(MediaType.APPLICATION_JSON)
                        .content(content)
            )
            .andDo(MockMvcResultHandlers.print())
            .andExpect(MockMvcResultMatchers.status().isOk())
            .andExpect(MockMvcResultMatchers.jsonPath("$").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].channel").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].channel.name").value(TEST_CHANNEL_01))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].channel.backend").value(Backend.SF_DATABUFFER.getKey()))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].pulseId").value(100))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].globalSeconds").value(
                  TestTimeUtils.getTimeStr(1, 0)))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].eventCount").value(5))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].value.min").value(100.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].value.mean").value(102.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].value.max").value(104.0))

            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].pulseId").value(105))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].globalSeconds").value(
                  TestTimeUtils.getTimeStr(1, 50000000)))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].eventCount").value(5))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].value.min").value(105.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].value.mean").value(107.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].value.max").value(109.0));
   }

   @Test
   public void testDateRangeQueryNrOfBinsAggregateExtrema() throws Exception {
      long startTime = 1000;
      long endTime = 1099;
      String startDate = TimeUtils.format(startTime);
      String endDate = TimeUtils.format(endTime);
      DAQQuery request =
            new DAQQuery(
                  new RequestRangeDate(
                        startDate,
                        endDate),
                  new AggregationDescriptor()
                        .setNrOfBins(2)
                        .setAggregations(Arrays.asList(Aggregation.min, Aggregation.mean, Aggregation.max))
                        .setExtrema(Arrays.asList(Extrema.minValue, Extrema.maxValue)),
                  TEST_CHANNEL_01);

      String content = mapper.writeValueAsString(request);
      System.out.println(content);

      this.mockMvc
            .perform(
                  MockMvcRequestBuilders
                        .post(QueryRestController.PATH_QUERY)
                        .contentType(MediaType.APPLICATION_JSON)
                        .content(content)
            )
            .andDo(MockMvcResultHandlers.print())
            .andExpect(MockMvcResultMatchers.status().isOk())
            .andExpect(MockMvcResultMatchers.jsonPath("$").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].channel").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].channel.name").value(TEST_CHANNEL_01))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].channel.backend").value(Backend.SF_DATABUFFER.getKey()))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].pulseId").value(100))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].globalSeconds").value(
                  TestTimeUtils.getTimeStr(1, 0)))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].eventCount").value(5))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].value.min").value(100.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].value.mean").value(102.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].value.max").value(104.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].extrema.minValue.value").value(100.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].extrema.minValue.pulseId").value(100))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].extrema.minValue.globalSeconds").value(
                  TestTimeUtils.getTimeStr(1, 0)))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].extrema.minValue.eventCount").value(1))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].extrema.maxValue.value").value(104.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].extrema.maxValue.pulseId").value(104))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].extrema.maxValue.globalSeconds").value(
                  TestTimeUtils.getTimeStr(1, 40000000)))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].extrema.maxValue.eventCount").value(1))

            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].pulseId").value(105))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].globalSeconds").value(
                  TestTimeUtils.getTimeStr(1, 50000000)))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].eventCount").value(5))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].value.min").value(105.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].value.mean").value(107.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].value.max").value(109.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].extrema.minValue.value").value(105.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].extrema.minValue.pulseId").value(105))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].extrema.minValue.globalSeconds").value(
                  TestTimeUtils.getTimeStr(1, 50000000)))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].extrema.minValue.eventCount").value(1))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].extrema.maxValue.value").value(109.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].extrema.maxValue.pulseId").value(109))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].extrema.maxValue.globalSeconds").value(
                  TestTimeUtils.getTimeStr(1, 90000000)))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].extrema.maxValue.eventCount").value(1));
   }

   @Test
   public void testDateRangeQueryBinSizeAggregate() throws Exception {
      long startTime = 10000;
      long endTime = 10999;
      String startDate = TimeUtils.format(startTime);
      String endDate = TimeUtils.format(endTime);
      DAQQuery request = new DAQQuery(
            new RequestRangeDate(
                  startDate,
                  endDate),
            new AggregationDescriptor().setDurationPerBin(100),
            TEST_CHANNEL_01);

      String content = mapper.writeValueAsString(request);
      System.out.println(content);

      this.mockMvc
            .perform(
                  MockMvcRequestBuilders
                        .post(QueryRestController.PATH_QUERY)
                        .contentType(MediaType.APPLICATION_JSON)
                        .content(content)
            )
            .andDo(MockMvcResultHandlers.print())
            .andExpect(MockMvcResultMatchers.status().isOk())
            .andExpect(MockMvcResultMatchers.jsonPath("$").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0]").exists())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].channel").isMap())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].channel.name").value(TEST_CHANNEL_01))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].channel.backend").value(Backend.SF_DATABUFFER.getKey()))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data").isArray())
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].pulseId").value(1000))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].globalSeconds").value(
                  TestTimeUtils.getTimeStr(10, 0)))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].eventCount").value(10))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].value.min").value(1000.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].value.mean").value(1004.5))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].value.max").value(1009.0))

            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].pulseId").value(1010))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].globalSeconds").value(
                  TestTimeUtils.getTimeStr(10, 100000000)))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].eventCount").value(10))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].value.min").value(1010.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].value.mean").value(1014.5))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].value.max").value(1019.0))

            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[2].pulseId").value(1020))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[2].globalSeconds").value(
                  TestTimeUtils.getTimeStr(10, 200000000)))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[2].eventCount").value(10))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[2].value.min").value(1020.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[2].value.mean").value(1024.5))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[2].value.max").value(1029.0))

            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[3].pulseId").value(1030))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[3].globalSeconds").value(
                  TestTimeUtils.getTimeStr(10, 300000000)))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[3].eventCount").value(10))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[3].value.min").value(1030.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[3].value.mean").value(1034.5))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[3].value.max").value(1039.0))

            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[4].pulseId").value(1040))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[4].globalSeconds").value(
                  TestTimeUtils.getTimeStr(10, 400000000)))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[4].eventCount").value(10))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[4].value.min").value(1040.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[4].value.mean").value(1044.5))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[4].value.max").value(1049.0))

            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[5].pulseId").value(1050))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[5].globalSeconds").value(
                  TestTimeUtils.getTimeStr(10, 500000000)))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[5].eventCount").value(10))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[5].value.min").value(1050.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[5].value.mean").value(1054.5))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[5].value.max").value(1059.0))

            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[6].pulseId").value(1060))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[6].globalSeconds").value(
                  TestTimeUtils.getTimeStr(10, 600000000)))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[6].eventCount").value(10))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[6].value.min").value(1060.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[6].value.mean").value(1064.5))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[6].value.max").value(1069.0))

            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[7].pulseId").value(1070))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[7].globalSeconds").value(
                  TestTimeUtils.getTimeStr(10, 700000000)))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[7].eventCount").value(10))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[7].value.min").value(1070.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[7].value.mean").value(1074.5))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[7].value.max").value(1079.0))

            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[8].pulseId").value(1080))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[8].globalSeconds").value(
                  TestTimeUtils.getTimeStr(10, 800000000)))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[8].eventCount").value(10))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[8].value.min").value(1080.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[8].value.mean").value(1084.5))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[8].value.max").value(1089.0))

            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[9].pulseId").value(1090))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[9].globalSeconds").value(
                  TestTimeUtils.getTimeStr(10, 900000000)))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[9].eventCount").value(10))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[9].value.min").value(1090.0))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[9].value.mean").value(1094.5))
            .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[9].value.max").value(1099.0));
   }

   @Test
   public void testDateRangeQueryIndexAggregate() throws Exception {
      long startTime = 1000;
      long endTime = 1099;
      String startDate = TimeUtils.format(startTime);
      String endDate = TimeUtils.format(endTime);
      DAQQuery request =
            new DAQQuery(
                  new RequestRangeDate(
                        startDate,
                        endDate),
                  new AggregationDescriptor()
                        .setAggregationType(AggregationType.index)
                        .setNrOfBins(2)
                        .setAggregations(Arrays.asList(Aggregation.min, Aggregation.mean, Aggregation.max)),
                  TEST_CHANNEL_WAVEFORM_01);

      String content = mapper.writeValueAsString(request);
      System.out.println(content);

      this.mockMvc
      .perform(
            MockMvcRequestBuilders
                  .post(QueryRestController.PATH_QUERY)
                  .contentType(MediaType.APPLICATION_JSON)
                  .content(content)
      )
      .andDo(MockMvcResultHandlers.print())
      .andExpect(MockMvcResultMatchers.status().isOk())
      .andExpect(MockMvcResultMatchers.jsonPath("$").isArray())
      .andExpect(MockMvcResultMatchers.jsonPath("$[0]").exists())
      .andExpect(MockMvcResultMatchers.jsonPath("$[0].channel").isMap())
      .andExpect(MockMvcResultMatchers.jsonPath("$[0].channel.name").value(TEST_CHANNEL_WAVEFORM_01))
      .andExpect(MockMvcResultMatchers.jsonPath("$[0].channel.backend").value(Backend.SF_DATABUFFER.getKey()))
      .andExpect(MockMvcResultMatchers.jsonPath("$[0].data").isArray())
      .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].pulseId").value(100))
      .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].globalSeconds").value(
            TestTimeUtils.getTimeStr(1, 0)))
      .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].eventCount").value(5))
      .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].value").isArray())
      .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].value[0].min").value(100.0))
      .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].value[0].mean").value(102.0))
      .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].value[0].max").value(104.0))
      .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].value[1].min").value(100.0))
      .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].value[1].mean").value(102.0))
      .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].value[1].max").value(104.0))

      .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].pulseId").value(105))
      .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].globalSeconds").value(
            TestTimeUtils.getTimeStr(1, 50000000)))
      .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].eventCount").value(5))
      .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].value").isArray())
      .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].value[0].min").value(105.0))
      .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].value[0].mean").value(107.0))
      .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].value[0].max").value(109.0))
      .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].value[1].min").value(105.0))
      .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].value[1].mean").value(107.0))
      .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].value[1].max").value(109.0));
   }

   @Test
   public void testDateRangeQueryIndexAggregateExtrema() throws Exception {
      long startTime = 1000;
      long endTime = 1099;
      String startDate = TimeUtils.format(startTime);
      String endDate = TimeUtils.format(endTime);
      DAQQuery request =
            new DAQQuery(
                  new RequestRangeDate(
                        startDate,
                        endDate),
                  new AggregationDescriptor()
                        .setAggregationType(AggregationType.index)
                        .setNrOfBins(2)
                        .setAggregations(Arrays.asList(Aggregation.min, Aggregation.mean, Aggregation.max))
                        .setExtrema(Arrays.asList(Extrema.minValue, Extrema.maxValue)),
                  TEST_CHANNEL_WAVEFORM_01);

      String content = mapper.writeValueAsString(request);
      System.out.println(content);

      this.mockMvc
      .perform(
            MockMvcRequestBuilders
                  .post(QueryRestController.PATH_QUERY)
                  .contentType(MediaType.APPLICATION_JSON)
                  .content(content)
      )
      .andDo(MockMvcResultHandlers.print())
      .andExpect(MockMvcResultMatchers.status().isOk())
      .andExpect(MockMvcResultMatchers.jsonPath("$").isArray())
      .andExpect(MockMvcResultMatchers.jsonPath("$[0]").exists())
      .andExpect(MockMvcResultMatchers.jsonPath("$[0].channel").isMap())
      .andExpect(MockMvcResultMatchers.jsonPath("$[0].channel.name").value(TEST_CHANNEL_WAVEFORM_01))
      .andExpect(MockMvcResultMatchers.jsonPath("$[0].channel.backend").value(Backend.SF_DATABUFFER.getKey()))
      .andExpect(MockMvcResultMatchers.jsonPath("$[0].data").isArray())
      .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].pulseId").value(100))
      .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].globalSeconds").value(
            TestTimeUtils.getTimeStr(1, 0)))
      .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].eventCount").value(5))
      .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].value").isArray())
      .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].value[0].min").value(100.0))
      .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].value[0].mean").value(102.0))
      .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].value[0].max").value(104.0))
      .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].value[1].min").value(100.0))
      .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].value[1].mean").value(102.0))
      .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].value[1].max").value(104.0))
      .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].extrema").isArray())
      .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].extrema[0].minValue.value").value(100.0))
      .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].extrema[0].minValue.pulseId").value(100))
      .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].extrema[0].minValue.globalSeconds").value(
            TestTimeUtils.getTimeStr(1, 0)))
      .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].extrema[0].minValue.eventCount").value(1))
      .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].extrema[0].maxValue.value").value(104.0))
      .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].extrema[0].maxValue.pulseId").value(104))
      .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].extrema[0].maxValue.globalSeconds").value(
            TestTimeUtils.getTimeStr(1, 40000000)))
      .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].extrema[0].maxValue.eventCount").value(1))
      .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].extrema[1].minValue.value").value(100.0))
      .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].extrema[1].minValue.pulseId").value(100))
      .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].extrema[1].minValue.globalSeconds").value(
            TestTimeUtils.getTimeStr(1, 0)))
      .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].extrema[1].minValue.eventCount").value(1))
      .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].extrema[1].maxValue.value").value(104.0))
      .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].extrema[1].maxValue.pulseId").value(104))
      .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].extrema[1].maxValue.globalSeconds").value(
            TestTimeUtils.getTimeStr(1, 40000000)))
      .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[0].extrema[1].maxValue.eventCount").value(1))

      .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].pulseId").value(105))
      .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].globalSeconds").value(
            TestTimeUtils.getTimeStr(1, 50000000)))
      .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].eventCount").value(5))
      .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].value").isArray())
      .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].value[0].min").value(105.0))
      .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].value[0].mean").value(107.0))
      .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].value[0].max").value(109.0))
      .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].value[1].min").value(105.0))
      .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].value[1].mean").value(107.0))
      .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].value[1].max").value(109.0))
      .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].extrema").isArray())
      .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].extrema[0].minValue.value").value(105.0))
      .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].extrema[0].minValue.pulseId").value(105))
      .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].extrema[0].minValue.globalSeconds").value(
            TestTimeUtils.getTimeStr(1, 50000000)))
      .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].extrema[0].minValue.eventCount").value(1))
      .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].extrema[0].maxValue.value").value(109.0))
      .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].extrema[0].maxValue.pulseId").value(109))
      .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].extrema[0].maxValue.globalSeconds").value(
            TestTimeUtils.getTimeStr(1, 90000000)))
      .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].extrema[0].maxValue.eventCount").value(1))
      .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].extrema[1].minValue.value").value(105.0))
      .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].extrema[1].minValue.pulseId").value(105))
      .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].extrema[1].minValue.globalSeconds").value(
            TestTimeUtils.getTimeStr(1, 50000000)))
      .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].extrema[1].minValue.eventCount").value(1))
      .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].extrema[1].maxValue.value").value(109.0))
      .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].extrema[1].maxValue.pulseId").value(109))
      .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].extrema[1].maxValue.globalSeconds").value(
            TestTimeUtils.getTimeStr(1, 90000000)))
      .andExpect(MockMvcResultMatchers.jsonPath("$[0].data[1].extrema[1].maxValue.eventCount").value(1));
   }

   @Test
   public void testGzipFileSuffixHeader() throws Exception {
      DAQQuery request = new DAQQuery(
            new RequestRangePulseId(
                  10,
                  11),
            TEST_CHANNEL_NAMES);
      request.setResponse(new JSONHTTPResponse(Compression.GZIP));

      String content = mapper.writeValueAsString(request);
      System.out.println(content);

      this.mockMvc
            .perform(MockMvcRequestBuilders
                  .post(QueryRestController.PATH_QUERY)
                  .contentType(MediaType.APPLICATION_JSON)
                  .content(content))

            .andDo(MockMvcResultHandlers.print())
            .andExpect(MockMvcResultMatchers.status().isOk())
            .andExpect(MockMvcResultMatchers.header().string("Content-Disposition", "attachment; filename=data.gz"));
   }

   @Test
   public void testJsonFileSuffixHeader() throws Exception {
      DAQQuery request = new DAQQuery(
            new RequestRangePulseId(
                  10,
                  11),
            TEST_CHANNEL_NAMES);

      String content = mapper.writeValueAsString(request);

      this.mockMvc
            .perform(MockMvcRequestBuilders
                  .post(QueryRestController.PATH_QUERY)
                  .contentType(MediaType.APPLICATION_JSON)
                  .content(content))

            .andDo(MockMvcResultHandlers.print())
            .andExpect(MockMvcResultMatchers.status().isOk())
            .andExpect(MockMvcResultMatchers.header().string("Content-Disposition", "attachment; filename=data.json"));
   }

   @Test
   public void testBadAggregation_01() throws Exception {
      DAQQuery request = new DAQQuery(
            new RequestRangePulseId(
                  10,
                  11),
            new AggregationDescriptor().setDurationPerBin(1000),
            TEST_CHANNEL_NAMES);

      String content = mapper.writeValueAsString(request);
      System.out.println(content);

      this.mockMvc
            .perform(MockMvcRequestBuilders
                  .post(QueryRestController.PATH_QUERY)
                  .contentType(MediaType.APPLICATION_JSON)
                  .content(content))

            .andDo(MockMvcResultHandlers.print())
            .andExpect(MockMvcResultMatchers.status().isBadRequest());
   }

   @Test
   public void testBadAggregation_02() throws Exception {
      DAQQuery request = new DAQQuery(
            new RequestRangeTime(
                  TimeUtils.getTimeFromMillis(0, 0),
                  TimeUtils.getTimeFromMillis(10, 0)),
            new AggregationDescriptor().setPulsesPerBin(100),
            TEST_CHANNEL_NAMES);

      String content = mapper.writeValueAsString(request);
      System.out.println(content);

      this.mockMvc
            .perform(MockMvcRequestBuilders
                  .post(QueryRestController.PATH_QUERY)
                  .contentType(MediaType.APPLICATION_JSON)
                  .content(content))

            .andDo(MockMvcResultHandlers.print())
            .andExpect(MockMvcResultMatchers.status().isBadRequest());
   }

   @Test
   public void testBadAggregation_03() throws Exception {
      DAQQuery request = new DAQQuery(
            new RequestRangePulseId(
                  10,
                  11),
            new AggregationDescriptor().setDurationPerBin(1000).setNrOfBins(100),
            TEST_CHANNEL_NAMES);

      String content = mapper.writeValueAsString(request);
      System.out.println(content);

      this.mockMvc
            .perform(MockMvcRequestBuilders
                  .post(QueryRestController.PATH_QUERY)
                  .contentType(MediaType.APPLICATION_JSON)
                  .content(content))

            .andDo(MockMvcResultHandlers.print())
            .andExpect(MockMvcResultMatchers.status().isBadRequest());
   }
}
